/*
 * Copyright (c) 2019 UPLEX Nils Goroll Systemoptimierung
 * All rights reserved
 *
 * Author: Geoffrey Simmons <geoffrey.simmons@uplex.de>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package vcl

import (
	"crypto/sha512"
	"encoding/binary"
	"hash"
	"math/big"
	"sort"
	"time"
)

const (
	RFC3339Micro = "2006-01-02T15:04:05.999999Z07:00"
)

// Meta represents a k8s object's Metadata
type Meta struct {
	Namespace       string
	UID             string
	ResourceVersion string
}

// Endpoint represents a Service endpoint.
type Endpoint struct {
	Meta
	IP   string
	Port int32
}

func (ep Endpoint) hash(hash hash.Hash) {
	portBytes := make([]byte, 4)
	binary.BigEndian.PutUint32(portBytes, uint32(ep.Port))
	hash.Write([]byte(ep.IP))
	hash.Write(portBytes)
}

// interface for sorting []Endpoint
type byIPPort []Endpoint

func (a byIPPort) Len() int      { return len(a) }
func (a byIPPort) Swap(i, j int) { a[i], a[j] = a[j], a[i] }
func (a byIPPort) Less(i, j int) bool {
	if a[i].IP < a[j].IP {
		return true
	}
	return a[i].Port < a[j].Port
}

// Service represents a backend Service, where the Endpoints are
// indexed by the Pod names to which they correspond.
type Service struct {
	Meta
	Endpoints map[string]Endpoint
}

func (svc Service) hash(hash hash.Hash) {
	for name, ep := range svc.Endpoints {
		hash.Write([]byte(name))
		ep.hash(hash)
	}
}

// ConfigMap represents metadata about the ConfigMap that contains
// VCL sources.
type ConfigMap struct {
	Meta
	Name    string
	Version string
}

// Spec is the specification for a VCL configuration derived from
// Services, Endpoints, and a ConfigMap representing VCL sources. This
// abstracts the VCL to be loaded by a Varnish instance.
type Spec struct {
	// Services are the Varnish backends, indexed by k8s Service
	// name.
	Services map[string]Service
	// CfgMap is the ConfigMap representing VCL sources.
	CfgMap ConfigMap
	// Main is the name of the main VCL source, which includes any
	// other sources, and is named as the file to load in a
	// vcl.load command.
	Main string
	// Version is the version string of the VCL config.
	Version string
	// Created is the time at which this Spec was created.
	Created time.Time
}

// DeepHash computes a alphanumerically encoded hash value from a Spec
// such that, almost certainly, two Specs are deeply equal iff their
// hash values are equal (unless we've discovered a SHA512 collision).
func (spec Spec) DeepHash() string {
	hash := sha512.New512_224()

	hash.Write([]byte(spec.Version))

	hash.Write([]byte(spec.CfgMap.Namespace))
	hash.Write([]byte(spec.CfgMap.Name))
	hash.Write([]byte(spec.CfgMap.UID))
	hash.Write([]byte(spec.CfgMap.ResourceVersion))

	svcs := make([]string, len(spec.Services))
	i := 0
	for k := range spec.Services {
		svcs[i] = k
		i++
	}
	sort.Strings(svcs)
	for _, svc := range svcs {
		hash.Write([]byte(svc))
		spec.Services[svc].hash(hash)
	}

	hash.Write([]byte(spec.Main))

	h := new(big.Int)
	h.SetBytes(hash.Sum(nil))
	return h.Text(62)
}

// FQVersion returns a fully qualified version -- the version string
// appended with the creation time of the Spec (since the same VCL
// version may be used with different backend configurations).
func (spec Spec) FQVersion() string {
	return spec.Version + "-" + spec.Created.Format(RFC3339Micro)
}

func (spec Spec) ConfigName() string {
	name := spec.FQVersion()
	return vclIllegal.ReplaceAllLiteralString(name, "_")
}
